(* ===================================================================

  This file provides:

  (1) destX functions which play the role of pattern matchers. 

       * destX   functions result in SOME / NONE

       * xdestX  functions result a value or throw an exception if 
                 the matching fail

  (2) a function to build the (more abstract) INSTR-tree out of 
      a HOL-term representation

  author: wishnu prasetya

=================================================================== *)

local 

open Parse HolKernel boolLib  
open integerTheory stringTheory
open Option Susp 
open Err
open L0Theory  L0_emb_mldata 

(* a variant of match_term *)

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun flip f (x,y) = f (y,x) ;

fun mk_binapp opr x y = mk_comb (mk_comb (opr,x), y) ;

fun mk_seq i j = mk_binapp (mk_const ("seq", Type `:INSTR->INSTR->INSTR`)) i j ;

fun mk_seqs s  = List.foldr (uncurry mk_seq) (Term `skip:INSTR`) s ;

(* get_argFunTy a->b should deliver b *)

fun get_argFunTy ty = List.nth (snd (dest_type ty),1)

(* a variant of delay *)

fun delay_ e = delay (fn () => e)

in 

(* ===================================================================

   The destX functions.

=================================================================== *)

fun destSkip t = 
    (if (fst o dest_const) t = "skip" then SOME () else NONE)
     handle
     _ => NONE

fun destPrint t = 
    (* a bit more complicated due to overloading on "print" *)
    case (matchTerm (Term `print (ikjsdakj:int)`) t,
          matchTerm (Term `print (ikjsdakj:string)`) t)
    of
       (SOME match, _ )   => (SOME o #residue o el 1) match

     | ( _ , SOME match ) => (SOME o #residue o el 1) match

     | (NONE,NONE)        => NONE
   ;

fun destAsg t = 
    case matchTerm (Term `jhgdsjg /:= ikjsdakj`) t of
       SOME match => 
            let
            val lhs = (#residue o el 1) match
            val rhs = (#residue o el 2) match
            in
            SOME (lhs,rhs)
            end
     | NONE => NONE
    ;


fun destSeq t = 
    case matchTerm (Term `seq jhgdsjg ikjsdakj`) t of
       SOME match =>
            let
            val i1 = (#residue o el 1) match
            val i2 = (#residue o el 2) match
            in
            case destSeq i2 of
                 SOME rest => SOME (i1 :: rest) 
               | NONE      => SOME [i1,i2]
            end
    |  NONE => NONE
    ;

fun destCond t =  SOME (dest_cond t) handle _ => NONE ;

fun destLET t = 
    let
    fun worker defrhss u =  
        let
        val (absbody,rhs) = dest_let u 
        in
        worker (rhs :: defrhss) absbody
        end
        handle _ => (* so, absbody is the innermost body *)
                    let
                    val (locvars,letbody) = strip_abs u
                    in
                    SOME (zip locvars defrhss, letbody)
                    end
    in
    if is_let t then worker [] t else NONE
    end ;

fun dest_tuple t 
     = 
     if pairLib.is_pair t 
        then let val (x,u) = pairLib.dest_pair t
             in
             (x :: dest_tuple u) 
             end
        else [t] ;


(* get a tuple of parameters, returning:

      1. a list of all parameters, in the same order as in the tuple;
         "REF" costructor will NOT be stripped.
   
      2. a list of pass-by-ref params; "REF" will be stripped 
 
      3. a list of pass-by-val params 
*)

fun dest_params t
    =
    if t = Term `void` then ([],[],[])
    else 
    let
    val params     =  dest_tuple t 
    fun isRefPar t =  isSome (matchTerm (Term `REF jkjakjh`) t)
    val RefPars    =  List.map rand (List.filter isRefPar params)
    val ValPars    =  List.filter (not o isRefPar) params
    in
    (params,RefPars, ValPars)
    end ;

fun destCall1 t = 
    case matchTerm (Term `/@ jhgdsjg hjgasj`) t of
         SOME match =>
              let
              val progname           = (#residue o el 1) match
              val (pars,Rpars,Vpars) = (dest_params o #residue o el 2) match
              in
              SOME (progname,pars,Rpars,Vpars)
              end
      |  NONE => NONE
    ;

fun destCall2 t = 
    case matchTerm (Term `bvcafagf /@= jhgdsjg hjgasj`) t of
         SOME match =>
              let
              val targetvar  = (#residue o el 1) match
              val progname   = (#residue o el 2) match
              val (pars,Rpars,Vpars) = (dest_params o #residue o el 3) match
              in
              SOME (targetvar,progname,pars,Rpars,Vpars)
              end
      |  NONE => NONE
    ;


fun destWhile t =
    case matchTerm (Term `wloop bvcafagf hjgasj`) t of

        SOME match =>
              let
              val guard   = (#residue o el 1) match
              val body    = (#residue o el 2) match

              val inv  = case matchTerm (Term `seq (assert bvcafagf) hjgasj`) body of
 
                           SOME match => (SOME o #residue o el 1) match
                         | _          =>  NONE

              in
              SOME (guard, inv, body)
              end

     |  _  => NONE
    ; 

fun destAssert t = 
    case matchTerm (Term `assert bvcafagf`) t of

        SOME match => (SOME o #residue o el 1) match
    |   NONE       =>  NONE
    ;


fun xdestPbody t =
    let
    fun norm_ t = case matchTerm (Term `vldscript return jhgdsjg`) t of

        SOME match => let
                      val i = (#residue o el 1 ) match
                      val e = (#residue o el 2 ) match
                      val retty = type_of e
                      val retassign = mk_binapp (mk_const ("/:=", retty --> retty --> Type `:INSTR`))
                                                (mk_const ("ret", retty))
                                                 e
                      in
                      (mk_seqs [i, retassign], retty)
                      end

        | _  => raise (L0Exn { errorCode  = SOME "xdestPbody"
                                , errorInside = NONE
                                , errorWhere = SOME t
                                , errorWhy   = SOME "A program-body should also specify a return value."
                                })

    fun normalize t = fst (norm_ t)

    fun get_retty t = snd (norm_ t)

    in
    case destLET t of (* the body is a let-structure *)

    SOME (decls,body)  =>  let
                           val rhss    = List.map snd decls
                           val locvars = List.map fst decls
                           fun mk_abslist b = List.foldr mk_abs b locvars
                           val new_body     = mk_abslist (normalize body)
                           in
                           (List.foldr (flip mk_let) new_body rhss, get_retty body)
                           end

    | _  => (* the body is not a let-structure *)

             norm_ t
                                                   
    end ;


fun xdestProgDefRHS t =
    case  matchTerm (Term `progdef_rhs iukjhsca kjhasiu kjhsdakjh`) t of      

       SOME match =>
            let
            val pre    = (#residue o el 1 ) match
            val post   = (#residue o el 2 ) match
            val pbody  = (#residue o el 3 ) match
            val (body,retty) = xdestPbody pbody
            in
            (pre,post,body,retty)
            end
     ;

fun xdestProgDef t =
    case matchTerm (Term `iukjhsca kjhsdakjh = jhgdsjg`) t of 

        SOME match =>
            let 
            val pname    = (#residue o el 1 ) match
            val (pars,rpars,vpars)  = (dest_params o #residue o el 2 ) match
            val rhs      = (xdestProgDefRHS o #residue o el 3 ) match
            in
            (pname,pars,rpars,vpars,rhs)
            end
   ;

(* ===================================================================

   Functions to build-up INSTR-tree

=================================================================== *)

fun buildINSTRtree t =
    (* SEQ *)
      case destSeq t of
         SOME z => SEQ { sself = delay_ t
                       , sbody = List.map  buildINSTRtree z
                       }
       | NONE   =>
    (* ASG *)
    case destAsg t of
         SOME (v,e) => ASG { aself = delay_ t, alhs = v, arhs = e }
       | NONE       =>
    (* print *)
    case destPrint t of
         SOME e => PRINT { pself = delay_ t, pexpr = e }
       | NONE   =>
    (* if-then-else *)
    case destCond t of
         SOME (g,i1,i2) => IF { iself = delay_ t
                              , iguard = g
                              , ithen  = buildINSTRtree i1
                              , ielse  = buildINSTRtree i2 }
       | NONE =>
    (* while *)
    case destWhile t of
         SOME (g,inv,body) => WHILE { wself = delay_ t
                                    , wguard = g
                                    , winv = inv
                                    , wbody = buildINSTRtree body }
       | NONE =>
    (* call-1 *)
    case destCall1 t of
         SOME (pname,pars,rpars,vpars) => 
               let
               val retType = (hd o snd o dest_type 
                                 o get_argFunTy 
                                 o type_of)
                                   pname
               val dummytarget = mk_const("skip", retType)
               in
               CALL { cself    = delay_ t
                    , clhs     = dummytarget
                    , cpname   = pname 
                    , cparams  = pars
                    , crparams = rpars
                    , cvparams = vpars }
               end
        | NONE =>
    (* call-2 *)
    case  destCall2 t of
         SOME (tvar,pname,pars,rpars,vpars) =>                               
               CALL { cself    = delay_ t
                    , clhs     = tvar
                    , cpname   = pname
                    , cparams  = pars
                    , crparams = rpars
                    , cvparams = vpars }
       | NONE =>
    (* local-block *)
    case destLET t of
         SOME (lvarsdefs,body) =>  LOCAL { lself = delay_ t
                                         , lvars = lvarsdefs
                                         , lbody =  buildINSTRtree body }
       | NONE => 
    (* ASSERT *)
    case destAssert t of 
         SOME a => ASSERT  { aself = delay_ t, aexpr = a }
       | NONE =>
    (* SKIP *)
    case destSkip t of
         SOME _ => SEQ { sself = delay_ t , sbody = [] }

       | NONE   => raise (L0Exn { errorCode  = SOME "buildINSTRtree"
                                , errorInside = NONE
                                , errorWhere = SOME t
                                , errorWhy   = SOME "Not recognized as an L0 instruction."
                                })
    (* if nothing matches then ERROR *)
    ;

(* ===================================================================

   Functions to build-up a PROG-structure

=================================================================== *)


fun buildPROG t = 
   let
   val (pname,pars,rpars,vpars,(pre,post,body,retty)) =  xdestProgDef t
   in
   { pname   = pname
   , pparams = pars
   , prparams = rpars
   , pvparams = vpars
   , ppre    = pre
   , ppost   = post
   , pbody   = body
   , preturn = retty
     }
   end
   handle
      L0Exn e => raise (L0Exn e)
   |  _       => raise  (L0Exn { errorCode  = SOME "buildPROG"
                                , errorInside = NONE
                                , errorWhere = SOME t
                                , errorWhy   = SOME "Not a correct syntax for a program definition."
                                })
   ;

end ;
