(* =============================================================

   The specification of a translator from L0min to ML

============================================================= *)

local

open Parse HolKernel stringLib stringSyntax
open PP util_PP
open L0_emb_mldata Translator Translator_gen
open L0min_Syntax 

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun elem x []     = false
  | elem x (y::s) = x=y orelse elem x s

fun trimQuote s = String.substring (s, 1, String.size s - 2)
                  handle _ => ""


fun mk_ref t = 
    let
    val ty = type_of t
    in
    mk_comb (mk_const ("REF", ty --> Type `:REF`), t)
    end 

val not_available_msg = "unimplemented feature... report it pls."



(* --- these are the translation tables of:

           * L0min operators 
           * base-types (int,bool,etc)

*)
        
val binop_translation =
      [ (Term `($+): int->int->int`,      pp_binop "+")
      , (Term `($-): int->int->int`,      pp_binop "-")
      , (Term `($*): int->int->int`,      pp_binop "*")
      , (Term `($/): int->int->int`,      pp_binop "/")
      , (Term `($=): int->int->bool`,     pp_binop "=")
      , (Term `($=): bool->bool->bool`,   pp_binop "=")
      , (Term `($=): void->void->bool`,   pp_binop "=")
      , (Term `($=): string->string->bool`, pp_binop "=")
      , (Term `($>): int->int->bool`,     pp_binop "<")
      , (Term `($>=): int->int->bool`,    pp_binop "<=")
      , (Term `($<): int->int->bool`,     pp_binop "<")
      , (Term `($<=): int->int->bool`,    pp_binop "<=")
      , (Term `($/\): bool->bool->bool`,  pp_binop "andalso")
      , (Term `($\/): bool->bool->bool`,  pp_binop "orelse")
      , (Term `($==>): bool->bool->bool`, 
                let (* IMP is a bit more .. *)
                fun ppimp ppx ppy
                      =
                      pp_binop "orelse" (pp_unop "not" ppx) ppy
                in
                ppimp
                end )
      ]

val unnop_translation =
    [ (Term `($~):bool->bool`, pp_unop "not")
    , (Term `($~):int->int`  , pp_unop "~")
    ]

val basetype_translation =
    [ (Type `:bool`  , (fn pps => add_string pps "bool"))
    , (Type `:int`   , (fn pps => add_string pps "int"))
    , (Type `:void`  , (fn pps => add_string pps "unit"))
    , (Type `:string`, (fn pps => add_string pps "string"))
    ]

fun find_translation trans_table t 
    = 
    case List.find (fn (t',_) => t'=t) trans_table of

           SOME (_,translation)  => translation

(* --- translating base-type *)
fun trans_basetype hol_ty pps = find_translation basetype_translation hol_ty pps

(* --- translating expression *)
fun trans_expr e pps =
      let
      fun texpr e_ pps = case e_ of
           Const t  => (if t = Term `void`   then add_string pps "()"
                        else if t = Term `T` then add_string pps "true"  
                        else if t = Term `F` then add_string pps "false"
                        else (* then it is either an integer or string constant *)
                        add_string pps (term_to_string t))

         | Ident v  =>  ( add_string pps "(!"
                        ; (add_string pps o fst o dest_var) v
                        ; add_string pps ")" )

         | Binop (binop,x,y) =>  (find_translation 
                                       binop_translation 
                                       binop)
                                       (texpr x) (texpr y)
                                       pps

        | Unop  (unop,x)    =>   (find_translation 
                                       unnop_translation
                                       unop)
                                       (texpr x) 
                                       pps
     in
     texpr (build_simple_L0minExpr e) pps
     end

in

structure L0min2ml_spec : Translator =

   struct

   type Toption  = { topt_skipassert : bool } (* skipping assert or not *)
   val option    = ref { topt_skipassert = true }

   val translate_expr          = trans_expr
   val translate_assertion     = (fn _ => raise Fail not_available_msg) 
   val translate_assertion_ret = (fn _ => raise Fail not_available_msg) 

   val tAlgebra  =

     { fprint  = 
         let
         fun tprint _ e pps = if is_string_literal e
                              then ( add_string pps "print "
                                   ; (add_string pps o trimQuote o term_to_string) e
                                   )
                              else (* then e must be of type-int *)
                                   ( add_string pps "print "
                                   ; translate_expr e pps )
         in
         tprint end

     , fasg    = 
         let
         fun targetstring v  = if is_var v
                               then fst (dest_var v)
                               else (* then v must be "ret" *)
                                    fst (dest_const v)
               
         fun tasg _ v e pps =  ( (add_string pps o targetstring) v
                               ; add_string pps " := "
                               ; translate_expr e pps  )
         in
         tasg end

     , fseq    = 
         let
         fun tseq _ ts pps = pp_list (fn pps => add_string pps ";") 
                                     (fn pps => add_string pps "(")
                                     (fn pps => add_string pps ")")
                                     (fn pps => add_string pps "()") 
                                      ts
                                      pps 
         in
         tseq end 

     , fcall   = 
          let
          fun tcall _ v pname pars _ _ pps =
              let
              val isDummy_v = ((fst o dest_const) v = "skip")
                              handle _ => false

              fun isRefPar x = isSome (matchTerm (Term `REF kjkjuyhk`) x)

              fun ppParam x pps = if isRefPar x 
                                  then (add_string pps o fst o dest_var o rand) x
                                  else translate_expr x pps

              in
              ( if isDummy_v 
                   then ()
                   else ( (add_string pps o fst o dest_var) v
                        ; add_string pps " := " )
              ; (add_string pps o fst o dest_const) pname 
              ; pp_list (fn pps => add_string pps ",") 
                        (fn pps => add_string pps "(")
                        (fn pps => add_string pps ")")
                        (fn pps => add_string pps "()") 
                        (List.map ppParam pars)
                         pps 
              )
              end
          in
          tcall end

     , fif     = 
          let
          fun tif _ g t1 t2 pps = pp_ml_if (translate_expr g) t1 t2 pps                        
          in
          tif end

     , flocal  = 
          let
          fun tdecl (var,e) =  ( (fn pps => (add_string pps  o fst o dest_var) var)
                               , (fn pps => (add_string pps "ref "
                                             ; translate_expr e pps ))
                               )

          fun tlocal _ decls tbody pps
              =
              pp_ml_let (List.map tdecl decls) tbody pps

         in
         tlocal end

     , fwhile  = 
          let
          fun twhile _ g _ tbody pps = pp_ml_while (translate_expr g) tbody pps
          in 
          twhile end 

     , fassert =  (* assert-instruction is for verification purpose ;
                     so it will be skipped *)
          let
          fun tassert _ e pps 
              = 
              if #topt_skipassert (!option)
              
              then add_string pps "()"

              else pp_ml_if (translate_expr e)
                            (fn pps => add_string pps "()")
                            (fn pps => ( add_string pps "raise (Fail \"L0-assertion violated: "
                                       ; ( add_string pps 
                                         o String.translate (fn c => if c = #"\\"
                                                                     then "\\\\" else String.str c)
                                         o term_to_string ) 
                                           e
                                       ; add_string pps "\")" 
                                       )) 
                             pps 
          in
          tassert end 
     }                


   fun translate_prog pname pars rpars vpars retty pp_body pps
             = 
             let
             fun pp_param (vname,pp_ty) pps = ( add_string pps vname
                                           ; add_string pps " : "
                                           ; pp_ty pps )

 
             fun trans_param v pps
                   = 
                   case List.find (fn v' => v'=v) vpars of
 
                        SOME _ (* pass-by-val *) => 

                                pp_param (fst (dest_var  v) ^ "_" , 
                                         
                                          trans_basetype (type_of v)) 

                                          pps
        
                      | NONE   (* pass-by-ref *) =>

                                let

                                val v_         = rand v
                                fun pp_ty  pps = ( add_string pps "("
                                                 ; trans_basetype (type_of v_) pps
                                                 ; add_string pps ") ref" )
                                in
                                pp_param (fst (dest_var v_) , pp_ty) pps
                                end
                 
             fun mk_new_vpar v  = let
                                  val v_ = (fst o dest_var) v
                                  in
                                  ((fn pps => add_string pps v_ )
                                  ,(fn pps => ( add_string pps "ref "
                                              ; add_string pps (v_ ^ "_")))
                                  )
                                  end

             fun pp_retdecl retty  = 
                 let
                 val ty   = fst (dest_type retty)
                 val init = if ty="int" then "0"
                            else if ty="bool" then "false"
                            else if ty="string" then ""
                            else (* then ty=void *) "()"
                 in
                 ( (fn pps => add_string pps "ret")
                 , (fn pps => (add_string pps "ref "; add_string pps init))
                 )
                 end


             fun ppbody retty pps = pp_ml_let [pp_retdecl retty]
                                              (fn pps => ( pp_body pps
                                                         ; add_newline pps              
                                                         ; add_string pps "; (!ret)"))
                                               pps

             in
             ( add_string pps "fun "
             ; begin_block pps CONSISTENT 0
             ; (add_string pps o fst o dest_const) pname
             ; add_string pps " "
             ; pp_list (fn pps => add_string pps ",")
                       (fn pps => add_string pps "(")
                       (fn pps => add_string pps ")")
                       (fn pps => add_string pps "()")
                       (List.map trans_param pars)
                        pps
             ; add_newline pps
             ; add_string pps "="
             ; add_newline pps
             ; if null vpars then ppbody retty pps
               else pp_ml_let (List.map mk_new_vpar vpars) 
                              (ppbody retty) 
                               pps
             ; end_block pps )
             end

 


   end (* of struct *)


(* ===================================================================

  Now generating the coresponding ML-translator

=================================================================== *)

structure L0min2ML = build_TR(L0min2ml_spec) ;
 
end 

