(* ===================================================================

  This file provides:

  (1) destX functions which play the role of pattern matchers. 

       * destX   functions result in SOME / NONE

       * xdestX  functions result a value or throw an exception if 
                 the matching fail

  (2) a function to build the (more abstract) TEST-tree out of 
      a HOL-term representation

  (3) some syntax pre-processors:

        * one to rewrite: whitebox (/@ P(...))  --> /@ P(..)
        * another to rewrite: whitebox (/@ P(..)) --> the body of P

  author: wishnu prasetya

=================================================================== *)

local 

open Parse HolKernel boolLib
open Option Susp 
open Err util_hol Logic_util   
open L0_emb_basic
open TestTheory  Test_emb_mldata 

infixr   //  ;
infix    <-  ;

(* a variant of match_term *)

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun flip f (x,y) = f (y,x)

(* a variant of delay *)

fun delay_ e = delay (fn () => e)

fun isSkip t = ((fst o dest_const) t = "skip") handle _ => false

fun deleteSkips ts = List.filter (not o isSkip) ts

(* match patterns to speed-up *)

val match_pat1 =  Term `whitebox (/@ kjhkjah  kjhkjhhk)`
val match_pat2 =  Term `whitebox (uiyiyury /@= kjhkjah  kjhkjhhk)`

in

fun destLabel t =
      (let 
       val (binder,absbody) = dest_comb t
       val (label,body)     = dest_abs absbody

       in

       if (fst o dest_const) binder = "/::"

       then SOME (fst (dest_var label), body)

       else NONE
       end )

       handle _ => NONE

fun destTEST t = 

      case (matchTerm (Term `TEST ikjsdakj`) t) of

         SOME match => let
                       val i = (#residue o el 1) match
                       in 
                       case destLabel i of

                           SOME (label,i_) => SOME (label,i_)
                         | _               => SOME ("?",i)
                       end

      | _ => NONE


fun destSUITE t =

      case (matchTerm (Term `SUITE ikjsdakj`) t) of

         SOME match => 

              let
              val testset_ = (#residue o el 1) match

              val (label,testset) = case destLabel testset_ of

                                       SOME (label,s) => (label,s)
                                      | _             => ("?",testset_)

              val tests = case destSeq testset of
                               SOME s => deleteSkips s
                             | _      => [testset]
              in 
              SOME (label, tests)
              end

      | _  => NONE 
               
(* ===================================================================

  For building a TEST tree out of HOL term

=================================================================== *)


fun buildTEST t = 

    case destTEST t of

         SOME (l,i)  => TEST { tlabel = l
                             , tbody  = i
                             , tself  = delay_ t }

       | _  =>

    case destSUITE t of

         SOME (l,ts) => SUITE { slabel = l
                              , sself  = delay_ t
                              , sset   = List.map buildTEST ts }

       | _ => 

    (* if nothing matches then ERROR *) 

    raise (L0Exn { errorCode  = SOME "buildTEST"
                                , errorInside = NONE
                                , errorWhere = SOME t
                                , errorWhy   = SOME "Not recognized as a test."
                                })

(* ===================================================================

  Pre-processors to deal with syntax sugars introduced by Test

=================================================================== *)

fun is_whitebox t = ((fst o dest_const) t = "whitebox")
                    handle 
                    _ => false

fun remove_whitebox_CONV t =
    if (fst o dest_const o rator) t = "whitebox"
    then mk_thm ([], mk_eq(t, rand t))
    else NO_CONV t 

fun remove_whitebox t = (rand o concl o DEPTH_CONV remove_whitebox_CONV) t 



fun expand_whitebox_CONV preprocess_CONV t 
    = 
    let

    fun rm_ref t = if is_ref t then rand t else t

    fun mk_expansion target pname aparams =
             let

             val (_,_,frpars,fvpars,(_,_,body,retty)) 
                     = 
                     (  xdestProgDef 
                      o concl 
                      o CONV_RULE preprocess_CONV
                      o TY_INST_prog_def aparams
                      o SPEC_ALL
                      o simple_getSpec ) pname

             val arpars   = (List.map rm_ref o List.filter is_ref) aparams

             val avpars   =  List.filter (not o is_ref) aparams

             val body_  = case target of

                             SOME x => body <- ([x]//[mk_ret retty])
                                            <- (arpars//frpars) 

                            | _     => body <- (arpars//frpars)

             val abody = list_mk_abs (fvpars,body_) 

             in
             List.foldl (flip mk_let) abody avpars
             end 

    in
    (* case matchTerm (--`whitebox (/@ kjhkjah  kjhkjhhk)`--) t of *)
    case matchTerm match_pat1 t of 

        SOME match  => let
                       val target    = NONE             
                       val pname     = (fst o dest_const o #residue o el 1) match
                       val aparams   = (dest_tuple o #residue o el 2) match
                       val expansion = mk_expansion target pname aparams
                       in
                       mk_thm ([], mk_eq(t,expansion))
                       end

    | _  => (* case matchTerm (--`whitebox (uiyiyury /@= kjhkjah  kjhkjhhk)`--) t of *)
             case matchTerm match_pat2 t of 

        SOME match  => let
                       val target    = (SOME o #residue o el 1) match
                       val pname     = (fst o dest_const o #residue o el 2) match
                       val aparams   = (dest_tuple o #residue o el 3) match
                       val expansion = mk_expansion target pname aparams
                       in
                       mk_thm ([], mk_eq(t,expansion))
                       end

    | _  => NO_CONV t
    
    end
 
fun expand_whitebox preprocess_CONV t 
        = 
        ( rand o concl 
        o DEPTH_CONV (expand_whitebox_CONV preprocess_CONV)) 
          t 

fun expand_wbox t = expand_whitebox ALL_CONV t 

end